"""Contains a base model for deep learning models to inherit from and have additional
functionalities, such as summaries (like with Keras) and easier initialization.
"""
from __future__ import annotations

from abc import ABC
from collections.abc import Callable
import os
from typing import BinaryIO, IO, Tuple, Union

import torch
import torch.nn as nn


class BaseModel(nn.Module, ABC):
    """Creates a base class for Deep Learning models with nice features.

    The class allows any inheriting model to have printed summaries, and easy
    initialization of certain layers.
    """

    def __str__(self) -> str:
        """Returns the name of the current class."""
        return type(self).__qualname__

    def get_param_count(self) -> Tuple[int, int]:
        """Returns the number of parameters of the model.

        Returns (tuple):
            (The number of trainable params, The number of non-trainable params)
        """
        param_count_nogrd = 0
        param_count_grd = 0
        for param in self.parameters():
            if param.requires_grad:
                param_count_grd += param.size().numel()
            else:
                param_count_nogrd += param.size().numel()
        return param_count_grd, param_count_nogrd

    def summary(
        self, half: bool = False, printer: Callable[[str], None] = print
    ) -> None:
        """Logs some information about the neural network.
        Args:
            printer: The printing function to use.
        """
        layers_count = len(list(self.modules()))
        printer(f"Model {self} has {layers_count} layers.")
        param_grd, param_nogrd = self.get_param_count()
        param_total = param_grd + param_nogrd
        printer(f"-> Total number of parameters: {param_total:n}")
        printer(f"-> Trainable parameters:       {param_grd:n}")
        printer(f"-> Non-trainable parameters:   {param_nogrd:n}")
        approx_size = param_total * (2.0 if half else 4.0) * 10e-7
        printer(f"Uncompressed size of the weights: {approx_size:.1f}MB")

    def save(self, filename: Union[str, os.PathLike, BinaryIO, IO[bytes]]) -> None:
        """Saves the model"""
        torch.save(self, filename)

    def half(self) -> BaseModel:
        """Transforms all the weights of the model in half precision.

        Note: this function fixes an issue on BatchNorm being half precision.
            See: https://discuss.pytorch.org/t/training-with-half-precision/11815/2
        """
        super().half()
        for layer in self.modules():
            if isinstance(layer, nn.BatchNorm2d):
                layer.float()
        return self

    def initialize_kernels(
        self,
        initializer: Callable[[torch.Tensor], torch.Tensor],
        conv: bool = False,
        linear: bool = False,
        batchnorm: bool = False,
    ) -> None:
        """Initializes the chosen set of kernels of the model.

        Args:
            initializer: a function that will apply the weight initialization.
            conv: Will initialize the kernels of the convolutions.
            linear: Will initialize the kernels of the linear layers.
            batchnorm: Will initialize the kernels of the batch norm layers.
        """
        for layer in self.modules():
            if linear and isinstance(layer, nn.Linear):
                initializer(layer.weight)
                continue
            if conv and isinstance(layer, nn.Conv2d):
                initializer(layer.weight)
                continue
            if batchnorm and isinstance(layer, (nn.BatchNorm2d, nn.GroupNorm)):
                initializer(layer.weight)
                continue

    def initialize_biases(
        self,
        initializer: Callable[[torch.Tensor], torch.Tensor],
        conv: bool = False,
        linear: bool = False,
        batchnorm: bool = False,
    ) -> None:
        """Initializes the chosen set of biases of the model.

        Args:
            initializer: A function that will apply the weight initialization.
            conv: Will initialize the biases of the convolutions.
            linear: Will initialize the biases of the linear layers.
            batchnorm: Will initialize the biases of the batch norm layers.
            **kwargs: Extra arguments to pass to the initializer function.
        """
        for layer in self.modules():
            if layer.bias is None:
                continue
            if linear and isinstance(layer, nn.Linear):
                initializer(layer.bias)  # type: ignore
            if conv and isinstance(layer, nn.Conv2d):
                initializer(layer.bias)  # type: ignore
            if batchnorm and isinstance(layer, (nn.BatchNorm2d, nn.GroupNorm)):
                initializer(layer.bias)  # type: ignore



"""Checks for the classes and functions in tiramisu"""
from functools import partial
import os
import unittest
import warnings
import numpy as np
import skimage
import skimage.io as skio
import skimage.transform as sktr
import torch
from torch import nn
from torch import optim
import torch_unittest
from tiramisu import ModuleName, DEFAULT_MODULE_BANK
from tiramisu import DenseLayer, DenseBlock, DenseUNet

OUTPUT_FOLDER = "test/output"


class TestDenseLayer(torch_unittest.TestCase):
    """Checks for the DenseLayer class."""

    def test_structure(self) -> None:
        """Checks that the layers of DenseLayer are correct and in good order."""
        dense_layer = DenseLayer(
            in_channels=8,
            out_channels=16,
            module_bank=DEFAULT_MODULE_BANK,
            checkpoint=False,
        )
        modules = [type(module).__name__ for module in dense_layer.children()]
        self.assertListEqual(modules, ["BatchNorm2d", "ReLU", "Conv2d", "Dropout2d"])

    def test_run_evalmode(self) -> None:
        """Checks the forward pass of the DenseLayer in eval mode."""
        torch.random.manual_seed(0)
        in_channels = 8
        # Preparation
        data = torch.ones(size=(1, in_channels, 1, 1), dtype=torch.float32)
        dense_layer = DenseLayer(
            in_channels=in_channels,
            out_channels=16,
            module_bank=DEFAULT_MODULE_BANK,
            checkpoint=False,
        ).eval()  # Disables batchnorm and dropout
        # Calculation
        result = dense_layer(data)
        # Check
        conv2d = dense_layer.conv
        assert isinstance(conv2d, nn.Conv2d)
        # The result can be computed in closed form
        weights, biases = conv2d.weight[:, :, 1, 1].sum(1), conv2d.bias
        self.assertTensorAlmostEqual(result.ravel(), weights + biases, places=4)

    def test_run_trainmode(self, checkpoint: bool = False) -> None:
        """Checks the forward pass of the DenseLayer in train mode."""
        torch.random.manual_seed(0)
        in_channels, out_channels = 8, 16
        # Preparation
        data = torch.ones(size=(2, in_channels, 1, 1), dtype=torch.float32)
        dense_layer = DenseLayer(
            in_channels=in_channels,
            out_channels=out_channels,
            module_bank=DEFAULT_MODULE_BANK,
            checkpoint=checkpoint,
        )
        dropped = torch.zeros(out_channels, dtype=torch.bool)
        dropped[[3, 9, 13]] = True
        # Calculation
        result = dense_layer(data)
        # Check
        # Note that the batchnormalization makes data = 0.0
        conv2d, dropout = dense_layer.conv, dense_layer.dropout
        assert isinstance(conv2d, nn.Conv2d)
        assert isinstance(dropout, nn.Dropout2d)
        # The result can be computed in closed form
        biases = conv2d.bias
        assert isinstance(biases, torch.Tensor)
        biases_scaled = biases / (1.0 - dropout.p)  # Dropout rescales the tensor
        expected = torch.where(dropped, torch.zeros_like(biases_scaled), biases_scaled)
        self.assertTensorAlmostEqual(result[0].ravel(), expected, places=4)

    def test_run_trainmode_checkpoint(self) -> None:
        """Checks the forward pass of the DenseLayer in train mode with checkpointing."""
        self.test_run_trainmode(checkpoint=True)


class TestDenseBlock(torch_unittest.TestCase):
    """Checks for the DenseBlock class."""

    def test_structure(self) -> None:
        """Checks that the DenseLayers of the DenseBlock are consistent."""
        nb_layers = 4
        dense_layer = DenseBlock(
            in_channels=8,
            growth_rate=16,
            nb_layers=nb_layers,
            module_bank=DEFAULT_MODULE_BANK,
            skip_block=False,
            checkpoint=False,
            prefix="",
        )
        modules = [name for name, _ in dense_layer.named_children()]
        expected = [f"dense_layer_{i}" for i in range(1, 5)]
        self.assertListEqual(modules, expected)


def nchw2hwc(image: torch.Tensor) -> torch.Tensor:
    """Transforms an image from torch to regular format (NCHW to HWC)."""
    return np.transpose(image[0], (1, 2, 0))


def hwc2nchw(image: np.ndarray) -> np.ndarray:
    """Transforms an image from regular format to torch format (HWC to NCHW)."""
    return np.transpose(image, (2, 0, 1))[np.newaxis]


class TestDenseUNet(torch_unittest.TestCase):
    """Checks for the DenseUNet class."""

    def setUp(self) -> None:
        """Sets up the DenseUNet model with the settings from the example in readme."""
        torch.random.manual_seed(0)

        module_bank = DEFAULT_MODULE_BANK.copy()
        # Dropout
        module_bank[ModuleName.DROPOUT] = partial(nn.Dropout2d, p=0.2, inplace=True)
        # Every activation in the model is going to be a GELU (Gaussian Error Linear
        # Units function). GELU(x) = x * Φ(x)
        # See: https://pytorch.org/docs/stable/generated/torch.nn.GELU.html
        module_bank[ModuleName.ACTIVATION] = nn.GELU
        # Example for image reconstruction:
        module_bank[ModuleName.ACTIVATION_FINAL] = nn.Sigmoid
        # Example for segmentation:
        # module_bank[ModuleName.ACTIVATION_FINAL] = partial(nn.LogSoftmax, dim=1)
        # Example for regression (default):
        # module_bank[ModuleName.ACTIVATION_FINAL] = nn.Identity

        self.model = DenseUNet(
            in_channels=1,  # Grayscale
            out_channels=3,  # 3-channel output (RGB)
            init_conv_filters=8,  # Number of channels outputted by the 1st convolution
            structure=(
                [2, 2],  # Down blocks
                2,  # bottleneck layers
                [2, 2],  # Up blocks
            ),
            growth_rate=8,  # Growth rate of the DenseLayers
            compression=1.0,  # No compression
            early_transition=False,  # No early transition
            include_top=True,  # Includes last layer and activation
            checkpoint=False,  # No memory checkpointing
            module_bank=module_bank,  # Modules to use
        )
        # Creating an optimizer
        # assert isinstance(self.model.parameters(), torch.Tensor)
        self.optimizer = optim.SGD(self.model.parameters(), lr=1e-3)

        # Loading the test image
        try:
            image = skimage.img_as_float32(skio.imread("test/data/woman_in_pink.jpg"))
        except FileNotFoundError as e:
            raise Exception(
                "Please run ./download_mockdata.sh before running this test."
            ) from e
        image = sktr.rescale(
            image[230 : 230 + 512, 60 : 60 + 512],  # Taking an even sized patch.
            scale=0.25,  # Makes the image of size 128x128 for faster computations.
            multichannel=2,  # Specifies the image is RGB
        )
        image_grayscale = image.mean(axis=-1, keepdims=True)

        self.X = torch.tensor(hwc2nchw(image_grayscale), dtype=torch.float32)
        self.y_true = torch.tensor(hwc2nchw(image), dtype=torch.float32)

        self.loss = nn.BCELoss()

        # Running the model (single training step)
        self.optimizer.zero_grad()
        self.y_pred_before_step = self.model(self.X)
        # Loss function
        self.loss_before_step = self.loss(
            self.y_pred_before_step.flatten(start_dim=1),
            self.y_true.flatten(start_dim=1),
        )
        self.loss_before_step.backward()
        self.optimizer.step()

        self.y_pred_after_step = self.model(self.X)
        self.loss_after_step = self.loss(
            self.y_pred_after_step.flatten(start_dim=1),
            self.y_true.flatten(start_dim=1),
        )

        # The prediction of the neural network will be outputed here
        if not os.path.exists(OUTPUT_FOLDER):
            os.mkdir(OUTPUT_FOLDER)

        # Saving the images
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            skio.imsave(
                os.path.join(OUTPUT_FOLDER, "groundtruth.png"),
                skimage.img_as_ubyte(image),
            )
            skio.imsave(
                os.path.join(OUTPUT_FOLDER, "input.png"),
                skimage.img_as_ubyte(image_grayscale),
            )
            skio.imsave(
                os.path.join(OUTPUT_FOLDER, "prediction_before_step.png"),
                skimage.img_as_ubyte(nchw2hwc(self.y_pred_before_step.detach())),
            )
            skio.imsave(
                os.path.join(OUTPUT_FOLDER, "prediction_after_step.png"),
                skimage.img_as_ubyte(nchw2hwc(self.y_pred_after_step.detach())),
            )

    def test_structure_main(self) -> None:
        """Checks that the main structure of the model is consistent."""
        main_modules = [name for name, _ in self.model.named_children()]
        expected = [
            "conv_init",
            "layers_down",
            "transitions_down",
            "bottleneck",
            "transitions_up",
            "layers_up",
            "conv_final",
            "activation_final",
        ]
        self.assertListEqual(main_modules, expected)

    def test_loss_improvement(self) -> None:
        """Checks that the loss improved after training one step."""
        self.assertGreater(self.loss_before_step, self.loss_after_step)

    def test_gradients(self) -> None:
        """Checks that the gradients exist everywhere in the model."""
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                with self.subTest(name=name):
                    self.assertIsNotNone(param)
                    self.assertNotEqual(torch.sum(param.grad ** 2).item(), 0.0)

    def test_image101(self) -> None:
        """Checks for inputing images that do not have a power of 2 for size."""
        X = self.X[:, :, :101, :101]
        y_pred = self.model(X)
        self.assertEqual(X.size()[2:], y_pred.size()[2:])


if __name__ == "__main__":
    unittest.main()



from enum import Enum, auto
from functools import partial
from collections.abc import Callable
from typing import Dict
from torch import nn


class ModuleType(Enum):
    """Module names that define which layers to use in the Tiramisu model."""

    CONV = auto()  # Convolution operations
    CONV_INIT = auto()  # Initial (1st) conv. operation. Kernel size must be provided.
    CONV_FINAL = auto()  # Final convolution. 1x1 kernel and reduce output to C classes.
    BATCHNORM = auto()  # Batch normalization
    POOLING = auto()  # Pooling operation (must reduce input size by a factor of two)
    # Note: if the size is odd, round *up* to the closest integer.
    DROPOUT = auto()  # Dropout
    UPSAMPLE = auto()  # Upsampling operation (must be by a factor of two)
    ACTIVATION = auto()  # Activation function to use everywhere
    ACTIVATION_FINAL = auto()  # Act. function at the last layer (e.g.softmax)


ModuleBankType = Dict[ModuleType, Callable[..., nn.Module]]
"""Type of module banks"""

UPSAMPLE2D_NEAREST = lambda in_channels, out_channels, module_bank: nn.Sequential(
    nn.UpsamplingNearest2d(scale_factor=2),
    module_bank[ModuleType.CONV](
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=3,
        padding=1,
    ),
    module_bank[ModuleType.ACTIVATION](),
)
# Pixel shuffle see: https://arxiv.org/abs/1609.05158
UPSAMPLE2D_PIXELSHUFFLE = lambda in_channels, out_channels, module_bank: nn.Sequential(
    module_bank[ModuleType.CONV](
        in_channels=in_channels,
        out_channels=4 * out_channels,
        kernel_size=3,
        padding=1,
    ),
    module_bank[ModuleType.ACTIVATION](),
    nn.PixelShuffle(2),
)
UPSAMPLE2D_TRANPOSE = lambda in_channels, out_channels, module_bank: nn.Sequential(
    nn.ConvTranspose2d(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=3,
        stride=2,
        padding=0,
    ),
    module_bank[ModuleType.ACTIVATION](),
)
DEFAULT_MODULE_BANK: ModuleBankType = {
    ModuleType.CONV: nn.Conv2d,
    ModuleType.CONV_INIT: partial(nn.Conv2d, kernel_size=3, padding=1),
    ModuleType.CONV_FINAL: nn.Conv2d,
    ModuleType.BATCHNORM: nn.BatchNorm2d,
    ModuleType.POOLING: partial(nn.MaxPool2d, kernel_size=2, ceil_mode=True),
    ModuleType.DROPOUT: partial(nn.Dropout2d, p=0.2, inplace=True),
    ModuleType.UPSAMPLE: UPSAMPLE2D_NEAREST,
    ModuleType.ACTIVATION: partial(nn.ReLU, inplace=True),
    ModuleType.ACTIVATION_FINAL: nn.Identity,
}



"""Checks for the classes and functions in layers.blurpool"""
import unittest
import torch
from tiramisu.layers.blurpool import BlurPool2d
import torch_unittest


class TestBlurPool(torch_unittest.TestCase):
    """Tests the BlurPool2d main class."""

    def test_dirac(self) -> None:
        """Checks that the result of the BlurPool2d is right on an example."""
        # Preparation
        data = torch.zeros(size=(1, 1, 8, 8), dtype=torch.float32)
        data[:, :, 2:6, 2:6] = 1.0  # White square in the middle
        blurpool_layer = BlurPool2d(1)
        # Calculation
        result = blurpool_layer(data)
        # Check
        expected = torch.tensor(
            [
                [
                    [
                        [0.0625, 0.2500, 0.2500, 0.0625],
                        [0.2500, 1.0000, 1.0000, 0.2500],
                        [0.2500, 1.0000, 1.0000, 0.2500],
                        [0.0625, 0.2500, 0.2500, 0.0625],
                    ]
                ]
            ],
            dtype=torch.float32,
        )
        self.assertTensorAlmostEqual(result, expected)

    def test_dense(self) -> None:
        """Checks that the output of the BlurPool2d has the right size (w/ 4D tensors)."""
        in_channels = 8
        # Preparation
        data = torch.zeros(size=(16, in_channels, 256, 256), dtype=torch.float32)
        blurpool_layer = BlurPool2d(in_channels)
        # Calculation
        result = blurpool_layer(data)
        # Check
        self.assertEqual(result.shape, torch.Size((16, in_channels, 128, 128)))


if __name__ == "__main__":
    unittest.main()



"""Implementation of the Tiramisu Deep Neural Network.

The file contains the modules and main class to instantiate a Tiramisu model. The
model isn't the same as in the original paper and has various improvements, such as
checkpointing, different types of upsampling, etc.

TODO:
    * Add unit-tests.
    * Make the Neural Network work at any input size (currently only works with
      powers of 2).
    * Add a default bank and make NN work with 3-D data.
"""
from collections.abc import Callable
from typing import List, Tuple, Union

# Deep Learning imports
import torch
import torch.nn as nn
import torch.utils.checkpoint as cp

# Local imports
from .model import BaseModel
from .modulebank import ModuleType, ModuleBankType, DEFAULT_MODULE_BANK


def _denselayer_factory(
    norm: nn.Module,
    activation: nn.Module,
    conv: nn.Module,
) -> Callable[..., torch.Tensor]:
    """This returns a callback to implement checkpointing."""

    def bn_function(*inputs: torch.Tensor) -> torch.Tensor:
        x = torch.cat(inputs, dim=1)
        x = norm(x)
        x = activation(x)
        x = conv(x)
        return x

    return bn_function


class DenseLayer(nn.Module):
    """A complete BN-ReLU-Conv-DropOut layer."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        module_bank: ModuleBankType,
        checkpoint: bool = False,
    ):
        """A superlayer containing all the basic blocs together.

        Performs a BatchNorm-ReLU-Conv-DropOut set of operation. Each individual part
        can be replaced by modifying the bank entries. More details in DenseUNet class.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            module_bank (ModuleBankType): The layers to use for common operations.
            checkpoint (bool, optional): Memory checkpointing. Defaults to False.
        """
        super().__init__()
        self.checkpoint = checkpoint
        # Standard Tiramisu Layer (BN-ReLU-Conv-DropOut)
        self.add_module(
            "batchnorm",
            module_bank[ModuleType.BATCHNORM](num_features=in_channels),
        )
        self.add_module("relu", module_bank[ModuleType.ACTIVATION]())
        self.add_module(
            "conv",
            module_bank[ModuleType.CONV](
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=3,
                stride=1,
                padding=1,
            ),
        )
        self.add_module("dropout", module_bank[ModuleType.DROPOUT]())

    @staticmethod
    def any_requires_grad(x: List[torch.Tensor]) -> bool:
        """Returns True if any of the layers in x requires gradients."""
        return any(layer.requires_grad for layer in x)

    def forward(self, inp: Union[List[torch.Tensor], torch.Tensor]) -> torch.Tensor:
        """Computes the foward pass of the layer."""
        inputs = inp if isinstance(inp, list) else [inp]  # Ensures inputs is a list
        assert isinstance(self.batchnorm, nn.Module)
        assert isinstance(self.relu, nn.Module)
        assert isinstance(self.conv, nn.Module)
        assert isinstance(self.dropout, nn.Module)
        bn_function = _denselayer_factory(self.batchnorm, self.relu, self.conv)
        if self.checkpoint and self.any_requires_grad(inputs):
            x = cp.checkpoint(bn_function, *inputs)
        else:
            x = bn_function(*inputs)
        x = self.dropout(x)
        return x


class DenseBlock(nn.ModuleDict):
    """Implementation of the denseblock of the Tiramisu, with all the skip connections."""

    def __init__(
        self,
        in_channels: int,
        growth_rate: int,
        nb_layers: int,
        module_bank: ModuleBankType,
        skip_block: bool = False,
        checkpoint: bool = False,
        prefix: str = "",
    ):  # pylint: disable=too-many-arguments
        """DenseBlock

        Args:
            in_channels (int): Number of channels in the input tensor.
            growth_rate (int): Number of channels in each DenseLayer.
            nb_layers (int): Number of DenseLayers to use.
            module_bank (ModuleBankType): The layers to use for common operations.
            skip_block (bool, optional): Adds a skip-connection over the full
                DenseBlock. Defaults to False.
            checkpoint (bool, optional): Memory checkpointing. Defaults to False.
            prefix (str, optional): Name prefix (pref. in snake case). Defaults to "".
        """
        super().__init__()
        for i in range(nb_layers):
            self.add_module(
                f"{prefix}dense_layer_{i+1}",
                DenseLayer(
                    in_channels + i * growth_rate,
                    growth_rate,
                    module_bank,
                    checkpoint,
                ),
            )

        self.skip_block = skip_block

    def forward(self, inp: torch.Tensor) -> torch.Tensor:  # type: ignore # pylint: disable=arguments-differ
        """Computes the forward pass of the denseblock."""
        skip_connections = [inp]

        for _, layer in self.items():
            out = layer(skip_connections)
            skip_connections.append(out)

        if self.skip_block:
            # Returns all of the x's, except for the first x (inp), concatenated
            # As we are not supposed to have skip connections over the full dense block.
            # See original tiramisu paper for more details
            return torch.cat(skip_connections[1:], 1)

        # Returns all of the x's concatenated
        return torch.cat(skip_connections, 1)


class TransitionDown(nn.Sequential):
    """A layer with a convolution and some operation that divides the resolution by 2."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        module_bank: ModuleBankType,
    ):
        """Layer dedicated to reducing the output width and height by a factor of two.

        Note: The POOLING module must downsize the input by two-folds.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            module_bank (ModuleBankType): The layers to use for common operations.
        """
        super().__init__()
        self.add_module("batchnorm", module_bank[ModuleType.BATCHNORM](in_channels))
        self.add_module("relu", module_bank[ModuleType.ACTIVATION]())
        self.add_module(
            "conv",
            module_bank[ModuleType.CONV](
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
                stride=1,
                padding=0,
            ),
        )
        self.add_module("dropout", module_bank[ModuleType.DROPOUT]())
        self.add_module("pool", module_bank[ModuleType.POOLING]())


def center_crop(layer: torch.Tensor, max_height: int, max_width: int) -> torch.Tensor:
    """ Crops a given layer to a certain size by removing equal margins all around."""
    _, _, height, width = layer.size()
    xy1 = (width - max_width) // 2
    xy2 = (height - max_height) // 2
    return layer[:, :, xy2 : (xy2 + max_height), xy1 : (xy1 + max_width)]


def crop_asym_like(layer: torch.Tensor, like: torch.Tensor) -> torch.Tensor:
    """Crops the right and bottom part of a tensor to look like `like`."""
    _, _, height_old, width_old = layer.size()
    _, _, height_new, width_new = like.size()
    assert (height_old >= height_new) and (
        width_old >= width_new
    ), "The size of the input layer is smaller than the desired one."
    return layer[:, :, :height_new, :width_new]


class TransitionUp(nn.Module):
    """Layer that increases the spatial resolution by a factor of 2."""

    def __init__(
        self, in_channels: int, out_channels: int, module_bank: ModuleBankType
    ):
        """Layer dedicated to increasing the output width and height by a factor of two.

        Note: The UPSAMPLE module must increase the scale of the input by two-folds.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            module_bank (ModuleBankType): The layers to use for common operations.
        """
        super().__init__()
        self.upsampling_layer = module_bank[ModuleType.UPSAMPLE](
            in_channels, out_channels, module_bank
        )

    def forward(self, inp: torch.Tensor, skip: torch.Tensor) -> torch.Tensor:
        """Computes the forward pass of the layer."""
        out = self.upsampling_layer(inp)
        out = crop_asym_like(out, skip)
        out = torch.cat([skip, out], 1)
        return out


class Tiramisu(BaseModel):
    """Tiramisu architecture
    Paper: The One Hundred Layers Tiramisu: Fully Convolutional DenseNets for
    Semantic Segmentation
    URL: https://arxiv.org/pdf/1611.09326.pdf
    Notes:
        Coded with the help of https://github.com/bfortuner/pytorch_tiramisu
        MIT License - Copyright (c) 2018 Brendan Fortuner
        and the help of https://github.com/keras-team/keras-contrib
        MIT License - Copyright (c) 2017 Fariz Rahman
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int = 1,
        init_conv_filters: int = 48,
        structure: Tuple[List[int], int, List[int]] = (
            [4, 4, 4, 4, 4],  # Down blocks
            4,  # bottleneck layers
            [4, 4, 4, 4, 4],  # Up blocks
        ),
        growth_rate: int = 12,
        compression: float = 1.0,
        early_transition: bool = False,
        include_top: bool = True,
        checkpoint: bool = False,
        module_bank: ModuleBankType = None,
    ):  # pylint: disable=too-many-arguments,too-many-locals
        """Creates a Tiramisu/Fully Convolutional DenseNet Neural Network for
        dense image prediction (e.g. segmentatation or pixel regression).

        Args:
            in_channels (int): The number of channels of the input images
            out_channels (int, optional): The number of predicted channels.
                Defaults to 1.
            init_conv_filters (int, optional): The number of filters of the
                very first layer. Defaults to 48.
            structure (Tuple[List[int], int, List[int]], optional): The number
                of DenseBlocks that compose the U-Net-shaped model. The tuple
                corresponds to the number and size of each dense block in,
                respectively, the encoding part, the bottleneck, and the
                decoding part. Defaults to (
                    [4, 4, 4, 4, 4],  # Down blocks
                    4,  # bottleneck layers
                    [4, 4, 4, 4, 4],  # Up blocks
                ).
            growth_rate (int, optional): The rate at which the DenseBlocks
                layers grow. Defaults to 12.
            compression (float, optional): Optimization where each of the
                DenseBlocks channels are reduced by a factor between 0 and 100%.
                Defaults to 1.0 (no compression).
            early_transition (bool, optional): Optimization where the input is
                downscaled by a factor of two after the first layer without
                skip-connection. Reduces memory usage. Defaults to False.
            include_top (bool, optional): Includes the final convolution and
                activation function. Defaults to True.
            checkpoint (bool, optional): Activates memory checkpointing, which
                reduces memory usage but increases latency. Defaults to False.
                See: https://arxiv.org/pdf/1707.06990.pdf
            module_bank (ModuleBankType, optional): The layers to use for
                common operations. Defaults to None (=DEFAULT_MODULE_BANK).
        """

        super().__init__()
        self.out_channels = out_channels
        self.init_conv_filters = init_conv_filters
        self.down_blocks = structure[0]
        self.bottleneck_layers = structure[1]
        self.up_blocks = structure[2]
        self.growth_rate = growth_rate
        self.compression = compression
        self.early_transition = early_transition
        self.include_top = include_top
        self.module_bank = module_bank or DEFAULT_MODULE_BANK

        channels_count = init_conv_filters
        # Keeps track of the number of channels for each skip-connection.
        skip_connections: List[int] = []

        # First layer
        self.conv_init = self.module_bank[ModuleType.CONV_INIT](
            in_channels=in_channels, out_channels=init_conv_filters
        )

        if early_transition:
            self.early_transition_down = TransitionDown(
                in_channels=channels_count,
                out_channels=int(channels_count * compression),
                module_bank=self.module_bank,
            )
            channels_count = int(channels_count * compression)

        # Downsampling part
        self.layers_down = nn.ModuleDict()
        self.transitions_down = nn.ModuleDict()
        for i, block_size in enumerate(self.down_blocks):
            self.layers_down.add_module(
                f"layers_down_{i}",
                DenseBlock(
                    in_channels=channels_count,
                    growth_rate=growth_rate,
                    nb_layers=block_size,
                    module_bank=self.module_bank,
                    skip_block=False,
                    checkpoint=checkpoint,
                    prefix=f"layers_down_{i}_",
                ),
            )
            channels_count += growth_rate * block_size
            skip_connections.insert(0, channels_count)
            self.transitions_down.add_module(
                f"transition_down_{i}",
                TransitionDown(
                    in_channels=channels_count,
                    out_channels=int(channels_count * compression),
                    module_bank=self.module_bank,
                ),
            )
            channels_count = int(channels_count * compression)

        # Bottleneck
        self.bottleneck = DenseBlock(
            in_channels=channels_count,
            growth_rate=growth_rate,
            nb_layers=self.bottleneck_layers,
            module_bank=self.module_bank,
            skip_block=True,
            checkpoint=checkpoint,
            prefix="bottleneck_",
        )
        prev_block_channels = growth_rate * self.bottleneck_layers
        channels_count += prev_block_channels

        # Upsampling part
        self.transitions_up = nn.ModuleDict()
        self.layers_up = nn.ModuleDict()
        for i, block_size in enumerate(self.up_blocks[:-1]):
            self.transitions_up.add_module(
                f"transition_up_{i}",
                TransitionUp(
                    in_channels=prev_block_channels,
                    out_channels=prev_block_channels,
                    module_bank=self.module_bank,
                ),
            )
            channels_count = prev_block_channels + skip_connections[i]
            self.layers_up.add_module(
                f"layers_up_{i}",
                DenseBlock(
                    in_channels=channels_count,
                    growth_rate=growth_rate,
                    nb_layers=block_size,
                    module_bank=self.module_bank,
                    skip_block=True,
                    checkpoint=checkpoint,
                    prefix=f"layers_up_{i}_",
                ),
            )
            prev_block_channels = growth_rate * block_size
            channels_count += prev_block_channels

        self.transitions_up.add_module(
            "transition_up_last",
            TransitionUp(
                in_channels=prev_block_channels,
                out_channels=prev_block_channels,
                module_bank=self.module_bank,
            ),
        )
        channels_count = prev_block_channels + skip_connections[-1]
        self.layers_up.add_module(
            "layers_up_last",
            DenseBlock(
                in_channels=channels_count,
                growth_rate=growth_rate,
                nb_layers=self.up_blocks[-1],
                module_bank=self.module_bank,
                skip_block=False,
                checkpoint=checkpoint,
                prefix="layers_up_last_",
            ),
        )
        channels_count += growth_rate * self.up_blocks[-1]

        if early_transition:
            self.early_transition_up = TransitionUp(
                in_channels=channels_count,
                out_channels=channels_count,
                module_bank=self.module_bank,
            )
            channels_count += init_conv_filters

        # Last layer
        if include_top:
            self.conv_final = self.module_bank[ModuleType.CONV_FINAL](
                in_channels=channels_count,
                out_channels=out_channels,
                kernel_size=1,
                stride=1,
                padding=0,
            )
            self.activation_final = self.module_bank[ModuleType.ACTIVATION_FINAL]()

    def forward(self, inp: torch.Tensor) -> torch.Tensor:
        """Computes the forward pass of the Tiramisu network."""
        x = self.conv_init(inp)

        transition_skip = None
        if self.early_transition:
            transition_skip = x
            x = self.early_transition_down(x)

        skip_connections = []
        for layer_down, transition_down in zip(
            self.layers_down.values(), self.transitions_down.values()
        ):
            x = layer_down(x)
            skip_connections.append(x)
            x = transition_down(x)

        x = self.bottleneck(x)

        for transition_up, layer_up in zip(
            self.transitions_up.values(), self.layers_up.values()
        ):
            skip = skip_connections.pop()
            x = transition_up(x, skip)
            x = layer_up(x)

        if self.early_transition:
            x = self.early_transition_up(x, skip=transition_skip)

        if self.include_top:
            # Computation of the final 1x1 convolution
            y_pred = self.conv_final(x)
            return self.activation_final(y_pred)

        return x

    def get_channels_count(self) -> List[int]:
        """Counts the number of out channels for each DenseBlocks and transitions.

        Returns: The list containing for each layers the number of (input) channels.
        """
        channels_count = [self.init_conv_filters]
        skip_connections: List[int] = []

        if self.early_transition:
            channels_count.append(int(channels_count[-1] * self.compression))

        # Downsampling part
        for block_size in self.down_blocks:
            channels_count.append(channels_count[-1] + self.growth_rate * block_size)
            skip_connections.insert(0, channels_count[-1])
            channels_count.append(int(channels_count[-1] * self.compression))

        # Bottleneck
        prev_block_channels = self.growth_rate * self.bottleneck_layers
        channels_count.append(channels_count[-1] + prev_block_channels)

        # Upsampling part
        for i, block_size in enumerate(self.up_blocks[:-1]):
            channels_count.append(prev_block_channels + skip_connections[i])
            prev_block_channels = self.growth_rate * block_size
            channels_count.append(channels_count[-1] + prev_block_channels)

        channels_count.append(prev_block_channels + skip_connections[-1])
        channels_count.append(
            channels_count[-1] + self.growth_rate * self.up_blocks[-1]
        )

        if self.early_transition:
            channels_count.append(channels_count[-1] + self.init_conv_filters)

        if self.include_top:
            channels_count.append(self.out_channels)

        return channels_count



"""Installation script for easy pip install."""
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="octopytorch",
    version="0.3",
    description="Segmentation Networks without a Backbone",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Nicolas Pielawski",
    author_email="nicolas@pielawski.fr",
    packages=setuptools.find_packages(),
    install_requires=["torch>=1.4"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""Blur Pooling layer to replace max pooling.

Blur pooling replaces the max pooling operation because this operation isn't completely
shift-invariant (the output varies in unexpected ways when the input tensor values are
shifted). The new operation consists of a regular max pooling followed by a blur
operation (gaussian kernel).

TODO:
    * Expand to 3-dimensional data.
    * Allow different filter sizes (now constrained to 3x3).
"""
import torch
import torch.nn as nn


class BlurPool2d(nn.Sequential):
    """Blur Pooling Layer (MaxPool2d replacement)

    Adds a blurring operation after the max pooling such that the layer is
    actually shift-invariant (invariance to small translations of the input).

    Official webpage: https://richzhang.github.io/antialiased-cnns/
    Paper: https://arxiv.org/abs/1904.11486

    Note:
        Improves shift-invariance, might remove artifacts (for segmentation) and might
        improve performance. Slower than regular max pooling.

    Arguments:
        in_features (int): The number of channels of the input tensor.
    """

    __constants__ = ["in_features"]
    _blur_kernel = torch.tensor(
        [[1 / 16, 2 / 16, 1 / 16], [2 / 16, 4 / 16, 2 / 16], [1 / 16, 2 / 16, 1 / 16]]
    )

    def __init__(self, in_features: int):
        """
        Args:
            in_features (int): The number of channels in the input
        """
        super().__init__()
        self.in_features = in_features

        self.add_module("maxpool", nn.MaxPool2d(2, stride=1))
        blurpool = nn.Conv2d(
            in_features,
            in_features,
            kernel_size=3,
            padding=1,
            stride=2,
            bias=False,
            groups=in_features,
        )
        blurpool.weight = torch.nn.Parameter(
            self._blur_kernel.repeat(in_features, 1, 1, 1), requires_grad=False
        )
        self.add_module("blurpool", blurpool)

    def extra_repr(self) -> str:
        return "in_features={}".format(self.in_features)
